package org.apache.tomee;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.security.Policy;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import org.apache.catalina.core.StandardContext;
import org.apache.catalina.webresources.StandardRoot;
import org.apache.openejb.AppModuleCreatorData;
import org.apache.openejb.OpenEJBCoreController;
import org.apache.openejb.OpenEJBException;
import org.apache.openejb.WebModuleCreatorData;
import org.apache.openejb.classloader.WebAppEnricher;
import org.apache.openejb.config.AppModule;
import org.apache.openejb.config.DeploymentLoader;
import org.apache.openejb.config.DeploymentModule;
import org.apache.openejb.config.WebModule;
import org.apache.openejb.jee.Application;
import org.apache.openejb.loader.SystemInstance;
import org.apache.openejb.sxc.ApplicationXml;
import org.apache.tomcat.vestige.launcher.VestigeWar;
import org.apache.tomcat.vestige.launcher.webresources.VestigeJarResourceSet;
import org.apache.tomee.catalina.CatalinaController;
import org.apache.xbean.finder.ResourceFinder;

import fr.gaellalire.vestige.spi.resolver.VestigeJar;
import fr.gaellalire.vestige.spi.resolver.VestigeJarEntry;

public class VestigeTomEEController implements OpenEJBCoreController, CatalinaController {

    private static final ThreadLocal<VestigeEar> VESTIGE_EAR_THREAD_LOCAL = new ThreadLocal<VestigeEar>();

    public void setResources(final StandardContext standardContext) {
        standardContext.setResources(new StandardRoot() {

            @Override
            public void processWebInfLib() {
                super.processWebInfLib();
                for (VestigeJar vestigeJar : webApp.vestigeWarDependencies) {
                    classResources.add(new VestigeJarResourceSet(this, "/WEB-INF/classes", vestigeJar, Collections.<VestigeJar> emptyList(), "/"));
                }
            }
        });
    }

    public String removeExtension(final String name) {
        if (name.endsWith(".vwar")) {
            return name.replaceFirst(".....$", "");
        }
        return null;
    }

    public Class<? extends DeploymentModule> getModuleClass(final File jarFile) {
        if (jarFile.getName().endsWith(".vear")) {
            return AppModule.class;
        } else if (jarFile.getName().endsWith(".vwar")) {
            return WebModule.class;
        }
        return null;
    }

    public WebModuleCreatorData createWebModule(final File warFile) {
        List<? extends VestigeJar> vestigeWarDependencies;
        WebModuleCreatorData webModuleCreatorData = null;
        VestigeEar vestigeEar = VESTIGE_EAR_THREAD_LOCAL.get();
        if (warFile.getName().endsWith(".vwar")) {
            webModuleCreatorData = new WebModuleCreatorData();
            try {
                final VestigeWar vestigeWar = VestigeWar.create(warFile, warFile.getName());

                VestigeJar vestigeWarJar = vestigeWar.getVestigeWar();
                // data/apps/myear/war1.vwar dynamically created ?

                // unpack in data/apps/mywar/WEB-INF
                // unpack in data/apps/myear/war1/WEB-INF
                warFile = unpack(vestigeWarJar.getFile(), warFile.getName(), VestigeWar.getUnpackDir());

                final Set<URL> webClassPath = new HashSet<URL>();
                final Set<URL> webRars = new HashSet<URL>();

                webClassPath.add(new File(new File(warFile, "WEB-INF"), "classes").toURI().toURL());

                vestigeWarDependencies = vestigeWar.getVestigeWarDependencies();
                for (VestigeJar vestigeJar : vestigeWarDependencies) {
                    if (vestigeJar.getName().endsWith(".rar")) {
                        webRars.add(vestigeJar.getFile().toURI().toURL());
                    } else {
                        webClassPath.add(vestigeJar.getFile().toURI().toURL());
                    }
                }

                final WebAppEnricher enricher = SystemInstance.get().getComponent(WebAppEnricher.class);
                if (enricher != null) {
                    webClassPath.addAll(Arrays.asList(enricher.enrichment(null)));
                }

                webModuleCreatorData.urls = new HashMap<String, URL[]>();
                webModuleCreatorData.urls.put(DeploymentLoader.URLS_KEY, webClassPath.toArray(new URL[webClassPath.size()]));
                webModuleCreatorData.urls.put(DeploymentLoader.RAR_URLS_KEY, webRars.toArray(new URL[webRars.size()]));
            } catch (Exception e) {
                throw new OpenEJBException(e);
            }
        } else if (vestigeEar != null) {
            webModuleCreatorData = new WebModuleCreatorData();
            try {
                // war inside vestige EAR
                vestigeWarDependencies = vestigeEar.getWarDependencies(warFile);

                warFile = unpack(warFile, warFile.getName(), VestigeWar.getUnpackDir());

                final Set<URL> webClassPath = new HashSet<URL>();
                final Set<URL> webRars = new HashSet<URL>();

                webClassPath.add(new File(new File(warFile, "WEB-INF"), "classes").toURI().toURL());

                for (VestigeJar vestigeJar : vestigeWarDependencies) {
                    if (vestigeJar.getName().endsWith(".rar")) {
                        webRars.add(vestigeJar.getFile().toURI().toURL());
                    } else {
                        webClassPath.add(vestigeJar.getFile().toURI().toURL());
                    }
                }

                final WebAppEnricher enricher = SystemInstance.get().getComponent(WebAppEnricher.class);
                if (enricher != null) {
                    webClassPath.addAll(Arrays.asList(enricher.enrichment(null)));
                }

                webModuleCreatorData.urls = new HashMap<String, URL[]>();
                webModuleCreatorData.urls.put(DeploymentLoader.URLS_KEY, webClassPath.toArray(new URL[webClassPath.size()]));
                webModuleCreatorData.urls.put(DeploymentLoader.RAR_URLS_KEY, webRars.toArray(new URL[webRars.size()]));
            } catch (Exception e) {
                throw new OpenEJBException(e);
            }
        } else {
            return null;
        }

    }

    @Override
    public AppModuleCreatorData createAppModule(final File jarFile) {
        if (!jarFile.getName().endsWith(".vear")) {
            return null;
        }
        Properties vestigeEARProperties = new Properties();

        AppModuleCreatorData appModuleCreatorData = new AppModuleCreatorData();
        try {
            VestigeEar vestigeEar = VestigeEar.create(jarFile);

            VestigeJar vestigeEarJar = vestigeEar.getVestigeJar();

            URL baseURL = new URL("jar", "", -1, vestigeEarJar.getFile().toURI().getPath() + "!/");
            VestigeJarEntry appXML = vestigeEarJar.getEntry("META-INF/application.xml");
            if (appXML != null) {
                appModuleCreatorData.application = ApplicationXml.unmarshal(appXML.open());
                appModuleCreatorData.applicationFound = true;
            } else {
                appModuleCreatorData.application = new Application();
            }

            VestigeJarEntry vestigeEAREntry = vestigeEarJar.getEntry("META-INF/maven-ear.xml");
            if (vestigeEAREntry != null) {
                vestigeEARProperties.load(vestigeEAREntry.open());
            }

            appModuleCreatorData.appDescriptors = new HashMap<String, URL>();

            for (String descriptor : DeploymentLoader.KNOWN_DESCRIPTORS) {
                VestigeJarEntry entry = vestigeEarJar.getEntry("META-INF/" + descriptor);
                if (entry != null) {
                    String name = entry.getName();
                    URL url = new URL(baseURL, name);
                    appModuleCreatorData.appDescriptors.put(name, url);
                }
            }

            File file = vestigeEarJar.getFile();
            appModuleCreatorData.appUrl = file.toURI().toURL();
            appModuleCreatorData.appId = file.getAbsolutePath();

            appModuleCreatorData.finder = new ResourceFinder() {
                @Override
                public URL find(final String uri) throws IOException {
                    return vestigeEar.getURL(uri);
                }
            };

        } catch (Exception e) {
            throw new OpenEJBException(e);
        }
        return appModuleCreatorData;
    }

    @Override
    public Policy getPolicy() {
        // TODO Auto-generated method stub
        return null;
    }

}
